import json
from lxml import etree
import chardet

from traspider.util import TraJson
from traspider.util.element.element import Element, ElementIter


class Response:
	def __init__(self, content, request, meta, response):
		self.content = content
		self.request = request
		self.meta = meta
		self.__response = response
	
	@property
	def status_code(self):
		return self.__response.status

	def json(self):
		json_data =  json.loads(self.content.decode())
		return TraJson(json_data)

	def text(self, encoding=None):
		"""自动解析html编码"""
		if encoding is None:
			char_code = chardet.detect(self.content)
			encoding = char_code.get('encoding', 'utf-8')
			if encoding is None:
				# 如果编码获取不到则抛出错误
				raise TypeError('<Unsupported encoding type, please specify encoding type.>')
		return self.content.decode(encoding, errors='ignore')

	def xpath(self, query):
		root = etree.HTML(self.text())
		xpath_result = root.xpath(query)
		if isinstance(xpath_result,list):
			return ElementIter(xpath_result)
		else:
			return Element(xpath_result)


	def __iter__(self):
		return (i for i in (self.status_code,self.request))

	def __repr__(self):
		class_name = type(self).__name__
		return "{}(method:{!r}, request:{!r})".format(
			class_name,
			*self
		)


from node_vm2 import VM

from traspider.util.exception import FileTypeUnsupported


class Node:
	def __init__(self, js_path=None):
		if js_path is None or js_path == "" or js_path.split(".")[-1] != "js":
			self.js_path = ""
		else:
			js_file = open(js_path, encoding="utf-8")
			self.js_path = js_file.read()

	def call_node(self, func_name, *args, **kwargs):
		if self.js_path == "":
			raise FileTypeUnsupported("<js file is empty or not a js file at all>")
		return self.__call_vm(func_name, *args, **kwargs)

	def __call_vm(self, func_name, *args, **kwargs):
		with VM() as vm:
			vm.run(self.js_path)
			return vm.call(func_name, *args, **kwargs)



import string

from traspider.util.trajson import ast


class Lexical:
	START_IDENTIFIER = set(string.ascii_letters + '_')
	VALID_IDENTIFIER = set(string.ascii_letters + string.digits + '_')
	VALID_NUMBER = set(string.digits)
	TOKENS = {
		'/': "line",
		"@": "flag",
		"*": "star",
		"]": "rbrackets",
	}

	def __init__(self, rule):
		self.__posit = 0
		self.__rule = rule
		self.__rule_len = len(self.__rule)
		self.__rule_list = list(self.__rule)
		self.__curr = self.__rule_list[self.__posit]
		self.__index = 0
		self.__lexical_word = list(self.lexical())

	def lexical(self):
		# 循环知道__curr为None
		while self.__curr is not None:
			# 如果_curr在TONKENS中
			if self.__curr in self.TOKENS:
				yield {'type': self.TOKENS[self.__curr],
					   'value': self.__curr,
					   'start': self.__posit, 'end': self.__posit + 1}
				self.__next()
			# 如果__curr 在 START_IDENTIFIER
			elif self.__curr in self.START_IDENTIFIER:
				# 记录当前字符的起始位置
				start = self.__posit
				buff = self.__curr
				while self.__next() in self.START_IDENTIFIER:
					buff += self.__curr
				yield {'type': 'unquoted_identifier', 'value': buff,
					   'start': start, 'end': start + len(buff)}
			elif self.__curr == "[":
				yield {'type': "lbrackets",
					   'value': self.__curr,
					   'start': self.__posit, 'end': self.__posit + 1}
				self.__next()
		yield  {'type': 'eof', 'value': '', 'start': self.__posit, 'end': self.__posit}

	def __next(self):
		"""
		返回规则列表中下一个元素
		:return:
		"""
		# 如果当前指针等于规则列表
		if self.__posit == self.__rule_len - 1:
			self.__curr = None
		else:
			# 将指针加1 返回下一个内容
			self.__posit += 1
			self.__curr = self.__rule_list[self.__posit]
		return self.__curr


	def convert_to_ast(self):
		init_word = self.__get_lexical()
		self.__add_index()
		nud_function = getattr(
			self, f'_type_{init_word["type"]}')
		left = nud_function(init_word["value"])

		curr_type = self.__get_lexical_type()
		next_lex_fun = getattr(
			self, f'_type_{curr_type}',None)
		if curr_type != "eof":
			if next_lex_fun is None:
				raise TypeError('<Attributes not supported for parsing>')
			else:
				self.__add_index()
				left = next_lex_fun(left)
				curr_type = self.__get_lexical_type()
		return left

	def __get_lexical(self,index=0):
		return self.__lexical_word[self.__index+index]

	def __get_lexical_type(self,index=0):
		return self.__lexical_word[self.__index+index]["type"]




	def __add_index(self):
		self.__index += 1


	def _type_unquoted_identifier(self,value):
		return ast.field(value)

	def _type_line(self,left):
		"""如果标签是line进入这个方法"""
		# 如果下一个标签是类型是不是star
		if not self.__get_lexical_type() == "star":
			right = self._parse_lin_rlbrack()
			if left['type'] == 'subexpression':
				left['children'].append(right)
				return left
			else:
				return ast.subexpression([left, right])

	def _parse_lin_rlbrack(self):
		type = self.__get_lexical_type()
		if type in ["unquoted_identifier", "star"]:
			return self.convert_to_ast()

	def _parse_arr_rlb(self):
		if self.__get_lexical_type() == "line":
			self.__verity_tag("line")
			right = self._parse_lin_rlbrack()
		return right


	def _type_lbrackets(self,left):
		self.__verity_tag('star')
		self.__verity_tag('rbrackets')
		right = self._parse_arr_rlb()
		return ast.arrexpression(left, right)

	def __verity_tag(self,tag):
		if self.__get_lexical_type() == tag:
			self.__add_index()
		else:
			# TODO 处理错误
			raise TypeError('<Syntax error, the list type only supports [*] expressions>')

class Visit:
	def __init__(self):
		self._method_cache = {}

	def visit(self,node,json_data):
		node_type = node['type']
		method = self._method_cache.get(node_type)
		if method is None:
			method = getattr(self,f"visit_{node_type}")
			self._method_cache[node_type] = method
		return method(node, json_data)

	def visit_subexpression(self, node, value):
		result = value
		for node in node['children']:
			result = self.visit(node, result)
		return result

	def visit_field(self, node, value):
		try:
			return value.get(node['value'])
		except AttributeError:
			return None

	def visit_arrexpression(self, node, value):
		base = self.visit(node['children'][0], value)
		if not isinstance(base, list):
			return None
		collected = []
		for element in base:
			current = self.visit(node['children'][1], element)
			if current is not None:
				collected.append(current)
		return collected





import os
import re
import sys

from traspider.core.engine import Engine
from traspider.core.request import Request
from traspider.core.spider import Spider
from traspider.util import exception, Encrypt
from traspider.util.node_vm.node import Node




sys.path.insert(0, re.sub(r"([\\/]items)|([\\/]spiders)", "", os.getcwd()))

__all__ = [
	"Node",
	"Spider",
	"Request",
	"Engine",
	"exception",
	"Node",
	"Encrypt"
]

__version__ = "0.0.07"



import asyncio
import json
import aiohttp
from traspider.util import Encrypt
from loguru import logger
from traspider.core.response import Response


class Download:
	def __init__(self, retry, time_out):

		self.retry = retry
		self.time_out = time_out
		self.encrypt = Encrypt()
		self.count = 0
		self.dedup = set()  # url去重
		self.error = dict()  # 重试链接
		self.error_count = 0

	async def download(self, spider, request):
		"""
		在这里可以处理下载前和下载后的处理
		:return:
		"""
		# 如果不是重试的request就放入到下载中间件中
		if not request.retry_:
			request = await spider.download_middleware(request)
		response = await self.crawl(request)
		# TODO 在这里做下载中间件之后的处理
		return response

	@logger.catch
	async def crawl(self, request):
		try:
			fingerprint_md5 = self.__encrypt_request(request)
			if self.__verify_request(fingerprint_md5):
				return
			async with aiohttp.ClientSession(
					headers=request.headers, connector=aiohttp.TCPConnector(ssl=False)
			) as session:
				if not self.error.get(fingerprint_md5):
					self.count += 1
				response = await session.request(
					method=request.method.upper(),
					url=request.url,
					params=request.params,
					data=request.data,
					proxy=request.proxy,
					proxy_auth=request.proxy_auth,
					timeout=self.time_out
				)
				logger.info(f"""<response: <Response {response.status}>> url:{request.url}
								请求次数:{self.error.get(fingerprint_md5, 0) + 1}""")
				self.dedup.add(self.__encrypt_request(request))
				return Response(content=await response.read(), request=request, meta=request.meta, response=response)
		except aiohttp.client_exceptions.ClientOSError as e:
			pass
		except aiohttp.client_exceptions.ClientConnectorError as e:
			pass
		except aiohttp.client_exceptions.ServerDisconnectedError as e:
			pass
		except asyncio.exceptions.TimeoutError as e:
			logger.error("请求超时!")
		except aiohttp.client_exceptions.ClientHttpProxyError as e:
			logger.error(f"代理链接失败:{e}")
		if not self.__error_retry(fingerprint_md5):
			logger.info(
				f"<重试请求次数:{self.error.get(fingerprint_md5, 0)}  url={request.url} params={request.params} data={request.data}>")
			request.retry_ = True
			return request

	def __error_retry(self, key):
		# 如果重试的request指纹次数等于重试次数
		if self.error.get(key) == self.retry:
			self.error_count += 1
			return True

		if self.error.get(key) is None:
			self.error[key] = 1
		else:
			self.error[key] += 1

	def __encrypt_request(self, request):
		"""
		对所有的请求进行md5放入到set中
		:param request:
		:return:
		"""
		data = request.data
		params = request.params
		if isinstance(data, dict):
			data = json.dumps(request.data)
		if data is None:
			data = ""
		if isinstance(params, dict):
			params = json.dumps(request.params)
		if params is None:
			params = ""
		return self.encrypt.md5(request.url + data + params)

	def __verify_request(self, fingerprint_md5):
		"""

		:param request:
		:return:
		"""
		return fingerprint_md5 in self.dedup



from lxml import etree

from traspider.util.traerror import ElementError


class Element():
    def __init__(self, element):
        self.element = element

    def extract(self):
        if isinstance(self.element, list):
            print(self.element)
            self.element = [element
                            if isinstance(element, str) else etree.tostring(
                element,
                method="html",
                encoding="unicode",
                with_tail=False
            ) for element in self.element]
        return self.element

    def extract_first(self):
        if isinstance(self.element, list) and len(self.element) > 0:
            # 如果self.element中第一位是字符串的话直接返回,否则将element对象转换为字符串
            return self.element[0] if isinstance(self.element[0], str) else etree.tostring(
                self.element[0],
                method="html",
                encoding="unicode",
                with_tail=False
            )
        if isinstance(self.element, str):
            return self.element
        raise ElementError("<element length is 0>")

    def get_index(self, index):
        if isinstance(self.element, list) and len(self.element) > 0:
            return self.element[index]
        if isinstance(self.element, str):
            raise ElementError("<Use get_index for strings>")
        raise ElementError("<element length is 0>")

    def xpath(self, query):
        xpath_list = self.element.xpath(query)
        if isinstance(xpath_list, list):
            return ElementIter(xpath_list)
        return Element(xpath_list)


class ElementIter(Element):

    def __init__(self, element):
        self.element = element

    def __iter__(self):
        for element in self.element:
            yield Element(element)

    def __repr__(self):
        return str(self.element)







def subexpression(children):
    return {"type": "subexpression", 'children': children}

def arrexpression(left, right):
    return {"type": "arrexpression", 'children': [left, right]}


def field(name):
    return {"type": "field", "children": [], "value": name}


import asyncio
import sys
import time

from loguru import logger
import csv

from traspider.util.log import Logging
from traspider.core.download import Download
from traspider.core.request import Request
from traspider.core.scheduler import Scheduler
from traspider.util.exception import FileTypeUnsupported
from traspider.util.tramysql.tramysql import AioMysql


class Engine:
	def __init__(self, spider):
		self.logging = Logging()
		self.spider = spider
		self.scheduler = Scheduler()
		self.download = Download(spider.retry,spider.time_out)
		self.item_count = 0
		self.loop = asyncio.get_event_loop()
		self.__task_list = []
		self.__loop_flag = True
		self.save_type = None
		self.aiomysql = AioMysql(self.spider.mysql_setting)
		self.__mysql_switch = False

	async def process_request(self, request):
		"""
		使用下载器下载请求
		:param request:
		:return:
		"""
		response = await self.download.download(self.spider, request)
		# 如果下载器下载失败response为空
		if response is None:
			return
		# 如果返回的是request类型则是重试的request  加入到队列中
		if isinstance(response, Request):
			await self.scheduler.add_scheduler(response)
			return
		await self.process_response(response, request)

	async def process_response(self, response, request):
		"""
		处理获取到的response并使用request的回调方法进行解析
		:param response:获取到的response
		:param request:发起请求的request
		:return:
		"""
		item_list = []
		# 如果callback为空的情况下,回调方法为默认的parse
		if request.callback is None:
			request.callback = self.spider.parse
		callback_results = request.callback(response, request)
		# 如果callback的结果为空再次不会
		if callback_results is None:
			return
		for result in callback_results:
			if isinstance(result, Request):
				await self.scheduler.add_scheduler(result)
			elif isinstance(result, dict):
				self.item_count += 1
				item_list.append(result)
			else:
				raise TypeError("item only supports dicts and lists")
		await self.process_item(item_list)

	async def process_future(self, future):
		self.__task_list.append(future)
		if len(self.__task_list) == self.spider.task_num or not await self.scheduler.scheduler_qsize():
			# 获取所有完成任务和未完成任务

			dones, pending = await asyncio.wait(self.__task_list)
			# 如果有未完成任务在此等待
			while pending:
				pass
			self.__loop_flag = False
			self.__task_list.clear()

	async def process_item(self, item):
		if len(item) == 0:
			return
		
		# 判断是否开启mysql
		if self.__mysql_switch:
			await self.aiomysql.batchInsert(item)
		if self.spider.save_type == "csv":
			asyncio.ensure_future(self.write_csv(item))
		elif self.spider.save_type == "txt":
			asyncio.ensure_future(self.write_txt(item))

	async def write_csv(self, item):
		with open(self.spider.save_path, "a+", newline="", encoding="utf-8") as f:
			csv_obj = csv.writer(f)
			csv_obj.writerows([i.values() for i in item])

	async def write_txt(self, items):
		with open(self.spider.save_path, "a+", newline="", encoding="utf-8") as f:
			write_data = [await self.__dict_to_str(item) for item in items]
			f.writelines(write_data)

	async def __dict_to_str(self, item):
		return " ".join(item.values()) + "\n"

	async def __init_mysql(self):
		# 如果mysql的配置不正确不运行
		conn_result = await self.aiomysql.inspection_conn()
		if conn_result is False:
			sys.exit()
		elif conn_result is None:
			self.__mysql_switch = False
		else:
			self.__mysql_switch = True

	async def _init_db(self):
		"""
		初始化数据库配置,进行链接检测
		:return:
		"""
		await self.__init_mysql()


	async def engine(self, start_requests):
		await self._init_db()
		for request in start_requests:
			await self.scheduler.add_scheduler(request)
		while self.__loop_flag or await self.scheduler.scheduler_qsize():
			request = await self.scheduler.next_request()
			future = asyncio.ensure_future(self.process_request(request))
			await self.process_future(future)

	def __print_log(self):
		print(
			"""

								   __                           _     __         
								  / /___  ________ __________  (_)___/ /__  _____
								 / __/  / ___/ __ `/ ___/ __ \/ / __  / _ \/ ___/
								/ /_/  / /  / /_/ (__  ) /_/ / / /_/ /  __/ /    
								\__/  /_/   \__,_/____/ .___/_/\__,_/\___/_/     
												  /_/                         
		"""
		)

	def start(self):
		self.__print_log()
		start = time.time()
		logger.info(f"{'*' * 20}爬虫启动{'*' * 20}")
		start_requests = iter(self.spider.start_request())
		self.loop.run_until_complete(self.loop.create_task(self.engine(start_requests)))
		logger.info(f"""spider end ...
								request_count:{self.download.count}
								error_request_count:{self.download.error_count}
								storage_item:{self.item_count}
								time_consuming:{time.time() - start}
								""")



from asyncio import Queue


class Scheduler:
	def __init__(self):
		self.queue = Queue()

	async def next_request(self):
		"""
		取出下一个请求的对象
		:return:
		"""
		return await self.queue.get()

	async def add_scheduler(self,request):
		"""
		将Request对象加入到调度器中
		:param request:
		:return:
		"""
		await self.queue.put(request)

	async def scheduler_qsize(self):
		"""
		查看当前调度器中的所有任务
		:return: 返回queue队列中的任务数
		"""
		return self.queue.qsize()



import json

from traspider.util.trajson.lexical import Lexical, Visit


class TraJson(dict):

	def xpath(self,xpath_str):
		if isinstance(xpath_str,str):
			lexical = Lexical(xpath_str)
			ast_obj = lexical.convert_to_ast()
			visit = Visit()
			return visit.visit(ast_obj,self)

















import copy
import json

from loguru import logger

from traspider import Engine

from traspider.core.request import Request
from traspider.core.response import Response
from traspider.util.exception import WrongParameter

from traspider.util.node_vm.node import Node


class Spider:
	def __init__(self):
		self.urls = []


	def start_request(self):
		for url in self.urls:
			if isinstance(url,str):
				yield Request(url, callback=self.parse)
			elif isinstance(url,dict):
				if url.get("url") is None:
					raise ValueError(f"The url attribute in urls cannot be empty:{url}")
				yield Request(method=url.get("method"),url=url.get("url"),params=url.get("params"),data=url.get("data"),callback=self.parse)
			else:
				raise ValueError("<Only dictionaries and strings can be stored in urls>")


	def parse(self, response: Response, request: Request):
		pass

	def download_middleware(self, request):
		"""爬虫下载器中间件"""
		return request

	def call_node(self, func, *args, **kwargs):
		return self.node.call_node(func, *args, **kwargs)

	def generate_total_Request(self, request=None, data=None, total=None, size=None, key=None):
		if not self.paging:
			return
		self.paging = False
		if isinstance(data, str) and request.url == data:
			for page in range(2, total):
				url_obj = request.parse_url(data)

				url_obj["query"][key] = page
				request.url = url_obj["url"]
				request.params = url_obj["query"]
				yield request
		else:
			if data is None:
				raise WrongParameter('<data cannot be None>')
			all_page = total // size if total % size == 0 else total // size + 1
			for page in range(2, all_page + 1):
				# 深拷贝request
				copy_request = copy.copy(request)
				# 如果request.data等于传入的data 或 loads后的data等于data
				if request.data == data or json.loads(request.data) == data:
					# 判断一下request.data的类型
					if isinstance(request.data,str):
						copy_request.data = json.loads(copy_request.data)
					copy_request.data[key] = page
				elif request.params == data:
					copy_request.params[key] = page
				else:
					raise WrongParameter('<Wrong parameter type, data can only be an attribute in the request>')
				yield copy_request

	@property
	def retry(self):
		if hasattr(self, "_Spider__retry"):
			return self.__retry
		return 100

	@retry.setter
	def retry(self, value):
		if value < 0 and not isinstance(value, int):
			raise ValueError('<The retry attribute must be an integer and greater than 0>')
		self.__retry = value

	@property
	def paging(self):
		if hasattr(self, "_Spider__paging"):
			return self.__paging
		return True

	@paging.setter
	def paging(self, value):
		self.__paging = value

	@property
	def mysql_setting(self):
		if hasattr(self, "_Spider__mysql_setting"):
			return self.__mysql_setting
		return {
			"host": "",
			"port": "",
			"user": "",
			"password": "",
			"db": "",
			"table": "",
		}

	@mysql_setting.setter
	def mysql_setting(self, value):
		if not isinstance(value, dict):
			raise ValueError('<mysql_setting must be a dictionary type>')
		if self.__check_mysql_setting_key(value):
			missing_keys = self.__check_mysql_setting_key(value)
			raise ValueError(f'<mysql_setting is missing the following fields:{missing_keys}>')
		self.__mysql_setting = value

	def __check_mysql_setting_key(self, mysql_setting: dict):
		"""
		检测mysql_setting中所需要的属性是否存在
		:param mysql_setting:
		:return:
		"""
		template = {
			"host": "",
			"port": "",
			"user": "",
			"password": "",
			"db": "",
			"table": "",
		}
		missing_keys = []
		for key in template.keys():
			if mysql_setting.get(key) is None:
				missing_keys.append(key)
		return missing_keys

	@property
	def save_path(self):
		if hasattr(self, "_Spider__save_path"):
			return self.__save_path
		return ""

	@save_path.setter
	def save_path(self, value: str):
		# 判断赋值路径是不是字符串
		if not isinstance(value, str):
			raise ValueError("<The save_path attribute only supports strings>")
		if value == "":
			self.__save_type = ""
			self.__save_path = value
			return
		if '.' not in value or len(value.split(".")) < 2:
			raise ValueError("<save_path is not a filename>")
		file_type = value.split(".")[-1]
		# 判断赋值的路径是不是csv或txt
		if file_type not in ["csv", "txt"]:
			raise ValueError(f'<{file_type} is an unsupported file type>')
		self.__save_type = file_type
		self.__save_path = value

	@property
	def save_type(self):
		if hasattr(self, "_Spider__save_type"):
			return self.__save_type
		return ""

	@property
	def node(self):
		if hasattr(self,"_Spider__node"):
			return self.__node
		return None

	@node.setter
	def node(self,value):
		if not isinstance(value,Node):
			raise ValueError('<The node attribute must be of Node class>')
		self.__node = value

	@property
	def task_num(self):
		if hasattr(self, "_Spider__task_num"):
			return self.__task_num
		return 100

	@task_num.setter
	def task_num(self,value):
		if value < 1 or value > 100:
			raise ValueError("task_num can only be between 1-100")
		self.__task_num = value

	@property
	def time_out(self):
		if hasattr(self, "_Spider__time_out"):
			return self.__time_out
		return 5

	@time_out.setter
	def time_out(self, value):
		if value < 1 or value > 20:
			raise ValueError("time_out can only be between 1-20")
		self.__time_out = value


	def start(self):
		engine = Engine(spider=self)
		engine.start()



import sys
from traspider.commands.create import CreateCommand



def _print_unknown_command(cmd_name):
    print("Unknown command: %s\n" % cmd_name)
    print('Output "traspider" to see how to create a spider file')


def _print_commands():
    print("Usage:")
    print("     traspider create -s <spider_name>")



def execute(argv=None):

    if argv is None:
        argv = sys.argv
    if len(argv) < 2:
        _print_commands()
        return

    cmd_name = argv.pop(1)
    cmd_list = {
        "create": CreateCommand
    }

    if not cmd_name:
        _print_commands()
        sys.exit(0)
    elif cmd_name not in cmd_list:
        _print_unknown_command(cmd_name)
        sys.exit(2)

    cmd = cmd_list[cmd_name]()
    cmd.add_arguments()
    cmd.run_cmd()

    sys.exit()


if __name__ == '__main__':
    execute()



import json
from urllib import parse

import aiohttp


class Request:
	def __init__(self, url, callback=None, method=None, headers=None, params=None, data=None, proxy=None, timeout=3,
				 meta={}):
		self.__proxy = proxy
		self.url = url
		self.method = method or "GET"
		self.headers = headers
		self.params = params
		self.__data = data
		self.timeout = timeout
		self.callback = callback
		self.meta = meta

	def parse_url(self, url=None):
		if url is None:
			url = self.url
		query = {i.split("=")[0]: i.split("=")[1] for i in parse.unquote(parse.urlsplit(url).query).split("&") if i}
		url = parse.urlparse(url)
		return {
			"url": url.scheme + "://" + url.netloc + url.path,
			"query": query
		}

	@property
	def data(self):
		return self.__data

	@data.setter
	def data(self, value):
		if value == "null":
			self.__data = None
		else:
			self.__data = value

	@property
	def retry_(self):
		if hasattr(self, "_Request__retry_"):
			return self.__retry_
		return False

	@retry_.setter
	def retry_(self, value):
		if isinstance(value, bool):
			self.__retry_ = value
		else:
			raise ValueError('<retry_ must be a bool type>')


	@property
	def proxy(self):
		if hasattr(self,"_Request__proxy"):
			return self.__proxy
		return None

	@proxy.setter
	def proxy(self,value):
		if isinstance(value,dict):
			if set(value.keys()) == set(["username","password","tunnel"]):
				self.proxy_auth = {
					"username":value["username"],
					"password":value["password"]
				}
				self.__proxy = value["tunnel"]
			else:
				raise ValueError('<Tunnel proxy must contain "username", "password", "tunnel">')
		elif isinstance(value,str):
			self.__proxy = value
		else:
			raise ValueError('<Proxy only supports dict and str>')

	@property
	def proxy_auth(self):
		if hasattr(self,"_Request__proxy_auth"):
			return self.__proxy_auth
		return None


	@proxy_auth.setter
	def proxy_auth(self,value):
		self.__proxy_auth = aiohttp.BasicAuth(value["username"], value["password"])

	def __iter__(self):
		return (i for i in
				(
					self.method, self.url, self.params, self.data, self.proxy, self.meta,
					getattr(self.callback, "__name__")
					if callable(self.callback)
					else self.callback))

	def __repr__(self):
		class_name = type(self).__name__
		return "{}(method:{!r}, url:{!r}, params:{!r}, data:{!r}, proxy:{!r}, meta:{!r}, callback:{!r})".format(
			class_name,
			*self
		)



import asyncio
import aiomysql
from loguru import logger


class AioMysql:

	def __init__(self, setting):
		self.setting = setting
		self.table_name = setting["table"]

	async def register(self, setting):
		'''
        初始化，获取数据库连接池
        :return:
        '''
		try:

			pool = await aiomysql.create_pool(host=setting["host"], port=setting["port"],
											  user=setting["user"], password=setting["password"],
											  db=setting["db"], autocommit=False)

			return pool
		except asyncio.CancelledError:
			raise asyncio.CancelledError
		except Exception as ex:
			logger.error("mysql数据库连接失败：{}".format(ex))

	async def get_curosr(self):
		'''
        获取db连接和cursor对象，用于db的读写操作
        :param pool:
        :return:
        '''
		pool = await self.register(self.setting)
		conn = await pool.acquire()
		cur = await conn.cursor()
		return conn, cur

	async def batchInsert(self, values):
		# 第一步获取连接和cursor对象
		conn, cur = await self.get_curosr()
		try:
			SQL = self.__generate_sql(values)
			# 将字典类型转为
			values = [self.__dict_to_tuple(item) for item in values]
			# 执行sql命令
			await cur.executemany(SQL, values)
			await conn.commit()
			logger.info(f"新增{cur.rowcount}条数据")
			# 返回sql执行后影响的行数
			return cur.rowcount
		finally:
			# 最后不能忘记释放掉连接，否则最终关闭连接池会有问题
			# await pool.release(conn)
			pass

	def __dict_to_tuple(self, item):
		if isinstance(item, dict):
			return tuple(item.values())
		else:
			raise TypeError("item type is not dict")

	def __generate_sql(self, values):
		"""
        生成sql语句
        :param values:
        :return:
        """
		if isinstance(values, list) and len(values) > 0:
			item = values[0]
		else:
			item = values
		sql = f"insert into {self.table_name}(" + self.__generate_key(
			item) + f") values ({','.join(['%s' for i in item.keys()])})"
		return sql

	def __generate_key(self, item):
		""" 生成key部分的sql语句"""
		sql = ""
		for index, key in enumerate(item.keys()):
			tag = ","
			if index == len(item.keys()) - 1:
				tag = ""
			sql += f"{key}{tag}"
		return sql

	async def inspection_conn(self):
		"""
		检测mysql是否链接正常
		:return:
		"""
		if 0 < list(self.setting.values()).count("") < 6:
			logger.error(f"mysql配置中有字段为空:{[key for key, val in self.setting.items() if val == '']}")
			return False
		elif list(self.setting.values()).count("") == 6:
			pass
		else:
			if not isinstance(self.setting["port"], int):
				logger.error("port字段必须是int类型")
				return False
			pool = await self.register(self.setting)
			if not pool is None:
				logger.info(f"{self.setting['host']}:{self.setting['port']}@{self.setting['db']}链接成功！")
				logger.info("mysql初始化正常")
				return True
			else:
				logger.error("mysql初始化失败")
				return False



import os
import re

from setuptools import find_packages, setup

_version_re = re.compile(r"__version__\s+=\s+(.*)")

def read_version():
    regexp = re.compile(r'^__version__\W*=\W*"([\d.abrc]+)"')
    init_py = os.path.join(os.path.dirname(__file__), "traspider", "__init__.py")
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)

def read(file_name):
    with open(
        os.path.join(os.path.dirname(__file__), file_name), mode="r", encoding="utf-8"
    ) as f:
        return f.read()



requires = [
            "aiohttp",
            "aiomysql",
            "aiosignal",
            "aiostream",
            "async-timeout",
            "asyncio",
            "attrs",
            "chardet",
            "charset-normalizer",
            "colorama",
            "frozenlist",
            "idna",
            "loguru",
            "lxml",
            "multidict",
            "node-vm2==0.3.5",
            "PyMySQL",
            "win32-setctime",
            "yarl",
        ]

setup(
    name="traspider",
    version=read_version(),
    author="NTrash",
    author_email='yinghui0214@163.com',
    description="An out-of-the-box lightweight asynchronous crawler framework",
    python_requires=">=3.7",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="MIT",
    install_requires=requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Development Status :: 3 - Alpha",
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
    entry_points={"console_scripts": ["traspider = traspider.commands.cmdline:execute"]},
)


import argparse
import os
import re
import shutil


class FullAction(argparse.Action):

	def __init__(self, option_strings, dest, default=None, required=False, help=None, metavar=None):
		super(FullAction, self).__init__(option_strings=option_strings, dest=dest, nargs=0, const=True,
										 default=default, required=required, help=help)

	def __call__(self, parser, namespace, values, option_string=None):
		if len(namespace.spider) > 0:
			setattr(namespace, self.dest, self.const)
		else:
			parser.error('must used with spider')


class CreateCommand:
	parser = None

	def add_arguments(self):
		parser = argparse.ArgumentParser(description="resolve")
		group = parser.add_mutually_exclusive_group()
		group.add_argument(
			"-s", "--spider", nargs=1, help="创建爬虫 如 traspider create -s <spider_name>"
		)
		self.parser = parser

	def create_spider(self, template_path, spider_name, project=None):
		spider_class_name = spider_name
		if spider_class_name.islower():
			spider_class_name = spider_class_name.title().replace("_", "")

		with open(template_path, "r", encoding="utf-8") as file:
			spider_template = file.read()
		spider_template = spider_template.replace("${spider_name}", spider_name)
		spider_template = spider_template.replace("${spider_class_name}", spider_class_name)
		result = self._save_spider_to_file(spider_template, spider_name, project)

		return result

	@staticmethod
	def copy_callback(src, dst, *, follow_symlinks=True):
		if src.endswith(".py"):
			with open(src, "r", encoding="utf-8") as src_file, open(
					dst, "w", encoding="utf8"
			) as dst_file:
				content = src_file.read()
				dst_file.write(content)
		else:
			shutil.copy2(src, dst, follow_symlinks=follow_symlinks)

	def run_cmd(self):
		args = self.parser.parse_args()
		if args.spider:
			spider_name = args.spider[0]
			create_type = "spider"
		else:
			raise
		# 检查spider_name
		if not re.search("^[a-zA-Z][a-zA-Z0-9_]*$", spider_name):
			raise Exception("命名不规范，请用下划线命名或驼峰命名方式")

		template_path = self._find_template(create_type)

		result = getattr(self, f"create_{create_type}")(template_path, spider_name)
		if result:
			print(f"{create_type} {spider_name} create success")

	@staticmethod
	def _find_template(template):
		if template == "spider":
			template_file = "spider_template.tmpl"

		template_path = os.path.abspath(
			os.path.join(__file__, "../../templates", template_file)
		)
		if os.path.exists(template_path):
			return template_path
		print("Unable to find template: %s\n" % template)

	def _save_spider_to_file(self, spider, spider_name, project_name):
		spider_underline = self._cover_to_underline(spider_name)
		if project_name:
			spider_file = f"{project_name}/{spider_underline}.py"
		else:
			spider_file = f"{spider_underline}.py"

		if os.path.exists(spider_file):
			print("文件已存在")
			return

		with open(spider_file, "w", encoding="utf-8") as file:
			file.write(spider)

		return True

	@staticmethod
	def _cover_to_underline(key):
		regex = "[A-Z]*"
		capitals = re.findall(regex, key)

		if capitals:
			for pos, capital in enumerate(capitals):
				if not capital:
					continue
				if pos == 0:
					if len(capital) > 1:
						key = key.replace(capital, capital.lower() + "_", 1)
					else:
						key = key.replace(capital, capital.lower(), 1)
				else:
					if len(capital) > 1:
						key = key.replace(capital, "_" + capital.lower() + "_", 1)
					else:
						key = key.replace(capital, "_" + capital.lower(), 1)

		return key
